# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient log module
"""
import logging as _logging

import urllib3

from coaclient import constants
from coaclient.cli import Arg, Actions


__all__ = (
    "add_logging",
    "LogLevels",
)


class LogLevels:
    """ Available levels of logging for coaclient """
    INFO = _logging.INFO
    DEBUG = _logging.DEBUG
    WARNING = _logging.WARNING
    ERROR = _logging.ERROR
    CRITICAL = _logging.CRITICAL
    LEVELS = [INFO, WARNING, ERROR, DEBUG, CRITICAL]

    @staticmethod
    def get_level_name(level):
        """
        Getting level name by level id or getting level id by level name
        """
        return _logging.getLevelName(level)

    @classmethod
    def log_levels(cls):
        """ Getting all available levels for logging """
        return [
            _logging.getLevelName(name) for name in cls.LEVELS
        ]


def add_logging(cli):
    """
    Add setup logging method to cli arguments methods
    Add additional arguments for logging configuration

    1. -l, --loglevel       Argument for setup loglevel for logger
                            Default: INFO
    2. --silence-urllib3    Argument which tell us should we show silence
                            urllib3 warnings or no.
    """
    cli.parser.defaults.update(setup_logging=setup_logging)
    # Add additional arguments for logging configuration
    cli.add_arguments(log_level=Arg(
        flags=('-l', '--log-level'),
        default=LogLevels.get_level_name(LogLevels.INFO),
        choices=LogLevels.log_levels(),
        help=constants.COURSERA_LOG_LEVEL_HELP
    ), silence_urllib3=Arg(
        flags=('--silence-urllib3',),
        action=Actions.STORE_TRUE,
        help=constants.COURSERA_LOG_SILENCE_URLLIB3
    ))
    # Setup these args as global for cli
    cli.parser.args.extend(['log_level', 'silence_urllib3'])


def setup_logging(args):
    """
    Computes and sets the logging configuration from the parsed arguments.
    """
    # Set CLI log format
    _logging.basicConfig(format="coaclient: %(message)s")
    # Get main logger
    logger = _logging.getLogger()
    # Setup log level for logger
    logger.setLevel(LogLevels.get_level_name(args.log_level))

    # Setup log level for urllib3 package logger
    _logging.getLogger("requests.packages.urllib3").setLevel(
        LogLevels.WARNING
    )
    if args.silence_urllib3 is True:
        # For details see:
        # https://urllib3.readthedocs.org/en/latest/security.html
        urllib3.disable_warnings()



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient CLI factory description
"""

from argparse import ArgumentParser
from dataclasses import asdict
from typing import List, Optional

from .args import Arg
from .parsers import Parser


class CLIFactory:
    """
    CLI Factory - factory for build CLI
    """

    def __init__(
            self,
            parser,
            *,
            args=None
    ):
        self._parser = parser
        self._args = args or {}

    @property
    def parser(self) -> Parser:
        """
        Get Parser description instance that uses as the main parser in CLI
        factory.
        """
        return self._parser

    def add_arguments(self, **args: Optional[Arg]) -> None:
        """
        Add arguments description to CLI factory which uses for the main
        parser and subparsers.
        """
        for arg_name, arg_conf in args.items():
            self._args[arg_name] = arg_conf

    def _setup_arguments(
            self,
            parser: ArgumentParser,
            args: List[str]
    ) -> None:
        for arg_name in args:
            arg_config = self._args.get(arg_name)
            if arg_config is None:
                continue
            kwargs = {}
            for field, value in asdict(arg_config).items():
                if field != "flags" and value is not None:
                    kwargs[field] = value
            parser.add_argument(*arg_config.flags, **kwargs)

    def _setup_subparsers(
            self,
            parser: ArgumentParser,
            parser_config: Parser
    ) -> None:
        subparsers = parser.add_subparsers(**parser_config.subparser.asdict())
        for p_config in parser_config.subparser.parsers:
            # Create sub parser
            _parser = subparsers.add_parser(p_config.name, **p_config.asdict(
                epilog=parser.epilog, description=parser.description
            ))
            # Setup parser arguments
            self._setup_arguments(_parser, p_config.args)
            # Setup default handler
            if p_config.func is not None:
                _parser.set_defaults(func=p_config.func)
            # Setup defaults
            if p_config.defaults:
                _parser.set_defaults(**p_config.defaults)
            # Setup subparsers
            if p_config.subparser is not None:
                self._setup_subparsers(_parser, p_config)

    def get_cli(self, *, args=None, **kwargs):
        """
        Build and return the main CLI parser.
        """
        # Create main parser
        parser = ArgumentParser(**self.parser.asdict(**kwargs))
        # Extend main parser args if provided
        self.parser.args.extend(args or [])
        # Setup arguments for main parser
        self._setup_arguments(parser, self.parser.args)
        # Setup defaults for main parser
        if self.parser.defaults:
            parser.set_defaults(**self.parser.defaults)

        if self.parser.subparser is not None:
            # Setup subparsers
            self._setup_subparsers(parser, self.parser)

        return parser



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth2.0 client library
"""
import os
from configparser import ConfigParser
from typing import Optional

from .settings import (
    OAUTH2_CONFIG_PATH,
    OAUTH2_CONFIG_FILE_NAME,
    OAUTH2_HOSTNAME,
    OAUTH2_PORT,
    OAUTH2_API_ENDPOINT,
    OAUTH2_AUTH_ENDPOINT,
    OAUTH2_TOKEN_ENDPOINT,
    OAUTH2_VERIFY_TLS,
    OAUTH2_TOKEN_CACHE_PATH,
)
from .utils import make_or_check_dir


class Config(ConfigParser):
    """ Configuration Coursera OAuth2.0 class """
    # _pylint: disable=too-many-ancestors
    # Disable this warning for Config or we can set the value of the
    # `max-parents` field to more than 7 in the .pylintrc file
    # (`max-parents` is 7 by default).
    OAUTH2_SECTION = "OAuth2"
    _FILE_PATH = os.path.join(
        OAUTH2_CONFIG_PATH, OAUTH2_CONFIG_FILE_NAME
    )

    @classmethod
    def load_from_file(cls, filename: Optional[str] = None):
        """
        Load configuration from file
        """
        config = cls()
        if filename is None:
            filename = cls._FILE_PATH
            make_or_check_dir(path=OAUTH2_CONFIG_PATH)
        cls._FILE_PATH = filename
        config.read([filename, ])
        config.set_default_values()
        return config

    def set_default_values(self, force: bool = False):
        """
        Set defaults values for the main OAuth2.0 section
        """
        if force is True or not super(Config, self).has_section(
                self.OAUTH2_SECTION
        ):
            self[self.OAUTH2_SECTION] = dict(
                hostname=OAUTH2_HOSTNAME,
                port=OAUTH2_PORT,
                api_endpoint=OAUTH2_API_ENDPOINT,
                auth_endpoint=OAUTH2_AUTH_ENDPOINT,
                token_endpoint=OAUTH2_TOKEN_ENDPOINT,
                verify_tls=OAUTH2_VERIFY_TLS,
                token_cache_path=OAUTH2_TOKEN_CACHE_PATH,
            )

    def save(self, filename: Optional[str] = None):
        """
        Save configuration to the file
        """
        filename = os.path.expanduser(filename or self._FILE_PATH)
        with open(filename, 'w') as file_descriptor:
            self.write(file_descriptor)

    def has_section(self, section: str) -> bool:
        """
        Check if section exist in configuration
        """
        if section == self.OAUTH2_SECTION:
            return False
        return super(Config, self).has_section(section)



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth2.0 client library
"""
__all__ = (
    "OAuth2ClientException",
    "OAuth2ConfigError",
    "OAuth2CacheException",
    "OAuth2TokenExpiredError",
)


class _OAuth2BaseException(Exception):
    """ Base exception class for custom exception """
    _STR = None

    def __init__(self, message: str, *args):
        self.message = message
        super(_OAuth2BaseException, self).__init__(message, *args)

    def __str__(self):
        if self._STR is None:
            return super(_OAuth2BaseException, self).__str__()
        return self._STR.format(
            message=self.message
        )


class OAuth2ClientException(_OAuth2BaseException):
    """ OAuth2.0 client exception class for custom exception """
    _STR = "Coursera OAuth2.0 client exception by OAuth2.0 protocol: {message}"


class OAuth2ConfigError(_OAuth2BaseException):
    """ Coursera OAuth2.0 Config custom error class """
    _STR = "Coursera OAuth2.0 configuration error: {message}"


class OAuth2CacheException(_OAuth2BaseException):
    """ Coursera OAuth2.0 Cache custom exception class """


class OAuth2TokenExpiredError(_OAuth2BaseException):
    """ OAuth2.0 token expired error class """
    _STR = "Coursera OAuth2.0 token expired error: {message}"



#!/usr/bin/env python3

# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Coursera's OAuth2 client library.
"""
import logging

from coaclient import (
    commands, constants
)
from coaclient.cli import (
    CLIFactory, Arg, Parser, SubParser
)
from coaclient.commands.exceptions import CoaClientCommandException
from coaclient.log import add_logging
from coaclient.oauth2.exceptions import (
    OAuth2ClientException,
    OAuth2ConfigError,
    OAuth2CacheException,
    OAuth2TokenExpiredError
)


def build_cli():
    """
    Build CLI for use Coursera OAuth2.0 client library

    return: CLIParser
    """
    cli_factory = CLIFactory(Parser(
        prog=constants.COURSERA_PROG_NAME, subparser=SubParser()
    ))
    # Initialize main used arguments
    cli_factory.add_arguments(
        config=Arg(
            flags=('-c', '--config'),
            type=str,
            help="Path to your custom the configuration file to use."
        )
    )
    cli_factory.parser.args.append('config')
    # Add logging configuration to Coursera CLI
    add_logging(cli_factory)

    # Add CLI commands
    commands.version.add_command(cli_factory)
    commands.config.add_command(cli_factory)

    return cli_factory.get_cli(
        description=constants.COURSERA_CLI_DESCRIPTION,
        epilog=constants.COURSERA_CLI_EPILOG
    )


def main():
    """
    Creating and initializing the CLI for processing input operations
    """
    cli = build_cli()
    args = cli.parse_args()
    # Configure logging
    args.setup_logging(args)
    logging.debug("[!!! DANGER !!!] Please do not use DEBUG mode for the "
                  "log file in production. Do it only on you own risk. "
                  "[!!! DANGER !!!]")

    status, message = 0, ""
    try:
        args.func(args)
    except AttributeError:
        status, message = (
            1, "Sub-command didn't set for processing your operation. "
               "Please see in `coaclient --help` how to use coaclient.\n"
        )
    except (
            CoaClientCommandException,
            OAuth2ClientException,
            OAuth2ConfigError,
            OAuth2CacheException,
            OAuth2TokenExpiredError
    ) as err:
        status, message = (1, "{err}\n".format(err=str(err)))
    finally:
        cli.exit(status, message)


if __name__ == "__main__":
    main()



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient CLI factory
"""

from .cli import CLIFactory
from .parsers import SubParser, Parser
from .args import Arg, Actions
from .formatters import RawTextArgsHelpFormatter

__all__ = (
    "CLIFactory",
    "Parser",
    "SubParser",
    "Arg",
    "Actions",
    # Help Formatters
    "RawTextArgsHelpFormatter"
)



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth2.0 client library
"""
import errno
import logging
import os
from http.server import BaseHTTPRequestHandler
from urllib.parse import urlparse, parse_qs

import status

from .exceptions import OAuth2CacheException, OAuth2ConfigError

__all__ = (
    "validate_input_data",
    "make_or_check_dir",
    "CourseraOAuth2CallbackHandler",
    "CallbackCodeHolder",
)


def validate_input_data(
        message: str,
        empty: bool = True,
        num_of_repeat: int = 3
):
    """
    Validation of input user data from CLI
    """
    if empty is True:
        return input(message).strip()

    counter = 0
    while True:
        value = input(message).strip()
        if not value:
            counter += 1
            logging.warning("Input value can't be empty. Please try again.")
        else:
            break
        if counter == num_of_repeat:
            raise OAuth2ConfigError(
                "Something wrong with your input data. Please look at input "
                "data and try again later."
            )
    return value


def make_or_check_dir(path: str):
    """
    Checking directory exist or no:
    1. if not exist creating it
    2. If exist check dir permissions
    """
    if not os.path.exists(path):
        try:
            os.makedirs(path, mode=0o700)
        except OSError as err:
            logging.debug(
                'Encountered an exception creating a directory for token '
                'cache file. Ignore it ...', exc_info=True
            )
            if err.errno != errno.EEXIST:
                raise
    else:
        permissions = oct(os.stat(path).st_mode)[-3:]
        if permissions != '700':
            raise OAuth2CacheException(
                "You have wrong permissions for token cache directory: "
                "{path}".format(path=path)
            )


class CallbackCodeHolder:
    """ A helper class to hold a token. """

    def __init__(self) -> None:
        self.code = None

    def __call__(self, code: str) -> None:
        self.code = code

    @property
    def exist(self) -> bool:
        """ Check if the code already stored in holder. """
        return self.code is not None


# Local Server handler to receive code from Coursera after authorization.
# That code used to get access tokens from Coursera to use Coursera API.
class CourseraOAuth2CallbackHandler(BaseHTTPRequestHandler):
    """ HTTPRequest handler class """
    _CALLBACK_PATH = "/callback"
    STATE = None
    CALLBACK = (lambda code: code)

    def _make_response(self, code: int, content: bytes):
        self.send_response(code)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(content)

    def _make_error_response(self, message):
        """ Return response with the error that raised """
        logging.warning('Return response with the error that raised:'
                        ' %s. %s', message, self.path)
        self._make_response(
            code=status.HTTP_400_BAD_REQUEST,
            content=message.encode("utf-8")
        )

    def do_GET(self):  # pylint: disable=invalid-name
        """ GET request processing """
        parsed = urlparse(self.path)

        if not parsed.query or parsed.path != self._CALLBACK_PATH:
            return self._make_error_response(
                "We encountered problems with your request."
            )

        params = parse_qs(parsed.query)
        # Checking state generated and received tokens
        if params.get('state') != [self.STATE, ]:
            return self._make_error_response(
                "State tokens didn't match. Please use last generated "
                "authorization URL to get access tokens."
            )

        code = params.get('code', [])
        if len(code) != 1:
            return self._make_error_response(
                "The wrong count of \"code\" values in query parameters."
            )

        if self.CALLBACK and callable(self.CALLBACK):
            self.CALLBACK(params['code'][0])

        return self._make_response(
            code=status.HTTP_200_OK,
            content="We have captured Coursera's response code. Feel "
                    "free to close this browser and come back to your "
                    "terminal. Thanks!".encode("utf-8")
        )



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient constants for CLI
"""

__all__ = (
    "COURSERA_PROG_NAME",
    "COURSERA_CLI_DESCRIPTION",
    "COURSERA_CLI_EPILOG",
    "COURSERA_LOG_LEVEL_HELP",
    "COURSERA_LOG_SILENCE_URLLIB3"
)

# Courser main cli constants
COURSERA_PROG_NAME = "coaclient"
COURSERA_CLI_DESCRIPTION = (
    "Coursera OAuth2 client CLI.\nThis tool helps users of the Coursera App "
    "Platform to programmatically access Coursera APIs."
)
COURSERA_CLI_EPILOG = (
    "Please file bugs on github at: "
    "https://github.com/blittlemore-coursera/coaclient/issues\n"
    "If you would like to contribute to this tool's development, "
    "check us out at: https://github.com/blittlemore-coursera/coaclient"
)

# Coursera logging constants
COURSERA_LOG_LEVEL_HELP = (
    "Used for setup log level for output messages.\n"
    "By default the log level is INFO."
)
COURSERA_LOG_SILENCE_URLLIB3 = (
    'Silence urllib3 warnings.\n'
    'See https://urllib3.readthedocs.org/en/latest/security.html for details.'
)



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth2.0 client library
"""
import logging
import os
import pickle
import subprocess
import sys
import time
import uuid
import webbrowser
from argparse import Namespace
from configparser import ConfigParser
from datetime import datetime
from http.server import HTTPServer
from typing import (
    Optional,
    Union,
    List,
    Dict,
    Any,
    Type
)

import requests
from requests.auth import AuthBase

from .config import Config
from .exceptions import (
    OAuth2ConfigError,
    OAuth2ClientException,
    OAuth2TokenExpiredError
)
from .settings import (
    OAUTH2_AUTH_ENDPOINT,
    OAUTH2_TOKEN_ENDPOINT,
    OAUTH2_VERIFY_TLS,
    OAUTH2_TOKEN_CACHE_PATH,
    OAUTH2_HOSTNAME,
    OAUTH2_PORT
)
from .utils import (
    make_or_check_dir,
    CallbackCodeHolder,
    CourseraOAuth2CallbackHandler,
    validate_input_data
)

logger = logging.getLogger()

__all__ = (
    "build",
)


def build(
        app_name: str,
        *,
        args: Optional[Namespace] = None,
        config: Optional[ConfigParser] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        scopes: Optional[Union[List[str], str]] = None,
        token_cache_file: Optional[str] = None,
        **kwargs
):
    """
    Creates a Coursera OAuth client based on the saved or provided
    configuration.
    """

    if app_name is None or not app_name:
        raise OAuth2ConfigError(
            "Please use the correct app name to create OAuth2.0 access to "
            "Coursera API."
        )

    if config is None:
        config_file = args.config if (
            args is not None and args.config is not None
        ) else None
        config = Config.load_from_file(filename=config_file)
        if not config.has_section(app_name):
            raise OAuth2ConfigError(
                "Please configure your App \"{app_name}\" using the "
                "`coaclient` cli tool or provide configuration via config "
                "file prior to use.".format(app_name=app_name)
            )

    return CourseraOAuth2(
        app_name,
        config=config,
        client_id=client_id,
        client_secret=client_secret,
        scopes=scopes,
        token_cache_file=token_cache_file,
        **kwargs
    )


class CourseraOAuth2Client(AuthBase):
    """ OAuth2.0 client for authorization in requests to Coursera API """

    def __init__(self, token: str, expires: float) -> None:
        self._token = token
        self._expires = expires

    def __call__(self, request):
        if self.is_valid:
            logging.debug("Adding an authorization header to the request.")
            request.headers["Authorization"] = "Bearer {token}".format(
                token=self._token
            )
            return request
        raise OAuth2TokenExpiredError("Expired at {expires}".format(
            expires=datetime.fromtimestamp(self._expires).strftime(
                "%d-%m-%Y %H:%M:%S"
            )
        ))

    @property
    def is_valid(self) -> bool:
        """
        OAuth2.0 client property.
        is_valid - return True if the token is not expired else False
        """
        return time.time() < self._expires


class CourseraOAuth2:
    """
    This class manages the OAuth2.0 tokens used to access Coursera's APIs.

    You must register your app with Coursera at:

        https://accounts.coursera.org/console

    Construct an instance of this class with the client_id and client_secret
    displayed in the Coursera app console. Please also set a redirect URL to be

        http://localhost:9876/callback

    Note: you can replace the port number (9876 above) with whatever port
    you'd like. If you would not like to use the local webserver to retrieve
    the codes set the local_port field in the constructor to None.
    """
    _TOKEN_TYPE = "BEARER"
    _MAC_OS = "darwin"
    _LINUX = "linux"

    def __init__(
            self,
            app_name: str,
            *,
            config: Optional[Config] = None,
            client_id: Optional[str] = None,
            client_secret: Optional[str] = None,
            scopes: Optional[Union[List[str], str]] = None,
            token_cache_file: Optional[str] = None,
            verify_tls: Optional[bool] = None,
            is_server_callback: bool = True,
            client_class: Type = CourseraOAuth2Client
    ) -> None:
        self.app_name = app_name
        self._config = config or Config.load_from_file()

        self._client_id = client_id or self._config.get(
            self.app_name, "client_id"
        )
        self._client_secret = client_secret or self._config.get(
            self.app_name, "client_secret"
        )
        self._scopes = scopes or self._config.get(self.app_name, "scopes")

        if isinstance(self._scopes, list):
            self._scopes = " ".join(self._scopes)

        self._auth_endpoint = self._config.get(
            self._config.OAUTH2_SECTION,
            "auth_endpoint",
            fallback=OAUTH2_AUTH_ENDPOINT
        )
        self._token_endpoint = self._config.get(
            self._config.OAUTH2_SECTION,
            "token_endpoint",
            fallback=OAUTH2_TOKEN_ENDPOINT
        )
        if verify_tls is not None and verify_tls in [False, True]:
            self._verify_tls = verify_tls
        else:
            self._verify_tls = self._config.getboolean(
                self._config.OAUTH2_SECTION,
                "verify_tls",
                fallback=OAUTH2_VERIFY_TLS
            )
        self._token_cache_path = self._config.get(
            self._config.OAUTH2_SECTION, "token_cache_path",
            fallback=OAUTH2_TOKEN_CACHE_PATH
        )
        self._token_cache_file = os.path.join(
            self._token_cache_path,
            token_cache_file or self._config.get(
                self.app_name, "token_cache_file"
            )
        )
        self._hostname = self._config.get(
            self._config.OAUTH2_SECTION, "hostname", fallback=OAUTH2_HOSTNAME
        )
        self._port = self._config.getint(
            self._config.OAUTH2_SECTION, "port", fallback=OAUTH2_PORT
        )

        self._is_server_callback = is_server_callback
        self._client_class = client_class
        # Cache token variable
        self._cache = None
        make_or_check_dir(path=self._token_cache_path)

    @property
    def cache(self) -> Dict[str, Any]:
        """
        Retrieve token from the file if the cache is empty and return it
        """
        if self._cache is None:
            self._cache = self._load_cache()
        return self._cache

    def _load_cache(self) -> Optional[Dict[str, Any]]:
        """ Reads the local file cache to get pre-authorized access tokens """
        try:
            logging.debug('Reading from local file cache: %s',
                          self._token_cache_file)
            with open(self._token_cache_file, 'rb') as file_descriptor:
                cache = pickle.load(file_descriptor)
                if self._cache_is_valid(cache):
                    logging.debug('Loaded from file system: %s', cache)
                else:
                    logging.warning('Unexpected value found in cache: %s', cache)
            return cache
        except IOError:
            logging.debug("The cache file doesn't exist in the file system: "
                          "%s", self._token_cache_file)
        except Exception as err:
            logging.exception('Unknown cache load exception detected: %s',
                              str(err), exc_info=True)

    @cache.setter
    def cache(self, cache: Dict[str, Any]):
        self._cache = cache
        self._save_cache(cache)

    def _save_cache(self, cache: Dict[str, Any]) -> None:
        """ Writes out OAuth2.0 tokens to the file a cache. """
        logging.debug('Writing to the local cache OAuth2.0 tokens.')
        if not self._cache_is_valid(cache):
            logging.error('Attempt to save invalid OAuth2 tokens: %s', cache)
            return

        try:
            logging.debug('Writing to cache file: %s', self._token_cache_file)
            with open(self._token_cache_file, 'wb') as file_descriptor:
                pickle.dump(cache, file_descriptor, protocol=-1)
                logging.debug('OAuth2.0 tokens successfully saved to '
                              'the cache file.')
        except Exception as err:
            logging.exception("Couldn't successfully cache OAuth2 tokens to "
                              "the cache file: %s", str(err), exc_info=True)

    @staticmethod
    def _cache_is_valid(cache: Dict[str, Any]) -> bool:
        """ Checks the cache for appropriate type correctness. """
        return (isinstance(cache, dict) and
                isinstance(cache.get("token"), str) and
                isinstance(cache.get("expires"), float) and
                (isinstance(cache.get("refresh"), str) or True))

    @property
    def _redirect_uri(self):
        return 'http://{hostname}:{port}/callback'.format(
            hostname=self._hostname,
            port=self._port,
        )

    def _build_auth_url(self, state: str) -> Optional[str]:
        auth = requests.Request(
            method='GET',
            url=self._auth_endpoint,
            params={
                'access_type': 'offline',
                'response_type': 'code',
                'client_id': self._client_id,
                'redirect_uri': self._redirect_uri,
                'scope': self._scopes,
                'state': state,
            }
        ).prepare()
        logging.debug('Constructed authorization request url: %s', auth.url)
        return auth.url

    def _auth_new_app(self):
        """
        Stands up a new localhost HTTP server and retrieves new OAuth2.0
        access tokens from the Coursera OAuth2.0 service.
        """
        logging.info('Requesting the new OAuth2.0 tokens from Coursera.')

        # Attempt to request new tokens from Coursera via the browser.
        state = uuid.uuid4().hex
        auth_url = self._build_auth_url(state)

        logging.info('Please visit the following URL to authorize this app.')
        logging.info(auth_url)
        logging.info('Look for additional details in the browser.')
        if sys.platform == self._MAC_OS:
            # OS X -- leverage the 'open' command present on all modern macs
            logging.info('Mac OS X detected; attempting to auto-open '
                         'the url in your default browser...')
            try:
                subprocess.check_call(['open', auth_url])
            except Exception as err:
                logging.exception('Could not call `open %s`. Exception: %s',
                                  auth_url, str(err))
        elif sys.platform == self._LINUX:
            logging.info('Linux detected; attempting to auto-open '
                         'the url in your default browser...')
            try:
                webbrowser.open(auth_url)
            except (TypeError, Exception) as err:
                logging.exception('Could not call `open %s`. Exception: %s',
                                  auth_url, str(err))

        if self._port is not None:
            # Boot up a local webserver to retrieve the response.
            code = CallbackCodeHolder()
            handler = CourseraOAuth2CallbackHandler
            handler.CALLBACK = code
            handler.STATE = state

            server = HTTPServer((self._hostname, self._port), handler)

            while not code.exist:
                server.handle_request()
            code = code.code
        else:
            code = validate_input_data(
                "Please enter the code received from Coursera: ", empty=False
            )

        return self._get_tokens_from_coursera({
            'code': code,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'redirect_uri': self._redirect_uri,
            'grant_type': 'authorization_code',
        })

    def _get_tokens_from_coursera(
            self, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Send the request to the Coursera token endpoint
        to get the new access token.
        """
        logging.debug('Send data %s to token endpoint %s',
                      data, self._token_endpoint)

        response = requests.post(
            url=self._token_endpoint, data=data,
            verify=self._verify_tls, timeout=10,
        )

        logging.debug('Response from token endpoint: (%s) %s',
                      response.status_code, response.text)

        if response.status_code != requests.codes.ok:  # pylint: disable=no-member
            logging.error(
                'Encountered unexpected status code. Status code: %s '
                'Response text: %s Response %s',
                response.status_code, response.text, response
            )
            raise OAuth2ClientException(
                'Unexpected status code from token endpoint. '
                'Status code: {status_code} '
                'Response text: {response_text}'.format(
                    status_code=response.status_code,
                    response_text=response.text
                )
            )
        # Parse JSON response data
        response_data = response.json()
        try:
            # Checking type of received token
            if response_data['token_type'].upper() != self._TOKEN_TYPE:
                logging.error('Unknown token type encountered in response '
                              'data: %s', response_data['token_type'])
                raise OAuth2ClientException(
                    'Unknown token type encountered in response data: '
                    '{token_type}'.format(
                        token_type=response_data['token_type']
                    )
                )
            tokens = {
                'token': response_data['access_token'],
                'expires': time.time() + response_data['expires_in']
            }
            refresh_token = response_data.get('refresh_token')
            if refresh_token is not None and isinstance(refresh_token, str):
                tokens['refresh'] = refresh_token
            return tokens
        except (KeyError, TypeError, AttributeError):
            logging.error('Some fields malformed or missing in the '
                          'response data. %s', response_data)
            raise OAuth2ClientException(
                'Some fields malformed or missing in the response data. '
                '{response_data}'.format(response_data=response_data)
            )

    def _exchange_refresh_tokens(self) -> Optional[Dict[str, Any]]:
        """
        Exchanges a refresh token for an access token
        """
        refresh_token = (self.cache or {}).get('refresh')
        tokens = None
        if refresh_token is not None:
            # Attempt to use the refresh token to get a new access token.
            tokens = self._get_tokens_from_coursera({
                'grant_type': 'refresh_token',
                'refresh_token': refresh_token,
                'client_id': self._client_id,
                'client_secret': self._client_secret,
            })
            tokens['refresh'] = refresh_token
        return tokens

    def _is_token_expired(self) -> bool:
        """
        The cache shouldn't be null and expired.
        """
        return self.cache is None or self.cache.get('expires', 0) < time.time()

    def refresh(self) -> bool:
        """
        Refresh tokens for access to Coursera API.
        """
        _tokens = self._exchange_refresh_tokens()
        if _tokens is not None:
            self.cache = _tokens
            return True
        return False

    @property
    def authorizer(self):
        """
        Checks cache, updates tokens if required and returns CourseraOAuth2Auth
        """
        if self._is_token_expired():
            logging.debug(
                "Attempting to use a refresh token to get new token."
            )
            _tokens = self._exchange_refresh_tokens()
            if _tokens is None:
                logging.info("Attempting to retrieve new tokens from the "
                             "auth endpoint. You will be prompted to "
                             "authorize your app in your web browser.")
                _tokens = self._auth_new_app()
                logging.debug("New received tokens: %s", _tokens)
            self.cache = _tokens
        else:
            logging.debug("Local cache with your tokens is good.")

        return self._client_class(self.cache.get('token'),
                                  self.cache.get('expires'))



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth2 client command

Command: version
"""

import sys

from coaclient import __version__
from coaclient.cli import Parser

from coaclient.log import LogLevels

__all__ = (
    "add_command",
)

_LOGS_LEVELS = [
    LogLevels.get_level_name(LogLevels.WARNING),
    LogLevels.get_level_name(LogLevels.ERROR)
]


def version(args):
    """
    Output the application version
    """
    msg = "Your {prog}'s version is: {version}s".format(
        prog=sys.argv[0].split('/')[-1], version=__version__
    )
    if args.log_level in _LOGS_LEVELS:
        msg = __version__
    print(msg)


def add_command(cli_factory):
    """
    Create version command with command handler and add to the Coursera's CLI
    """
    cli_factory.parser.subparser.parsers.append(Parser(
        name="version",
        help="Output the version %(prog)s.",
        func=version
    ))



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient Help formatter for CLI factory
"""

from argparse import (
    RawTextHelpFormatter,
    ArgumentDefaultsHelpFormatter
)

__all__ = (
    "RawTextArgsHelpFormatter",
)


class RawTextArgsHelpFormatter(RawTextHelpFormatter,
                               ArgumentDefaultsHelpFormatter):
    """
    RawTextArgsHelpFormatter - Formatter for print help information
    """

    def _format_action_invocation(self, action):
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar

        # if the Optional doesn't take a value, format is:
        #    -s, --long
        if action.nargs == 0:
            return ', '.join(action.option_strings)

        # if the Optional takes a value, format is:
        #    -s, --long ARGS
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return '%s %s' % (
            ', '.join(action.option_strings), args_string
        )



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient Parsers description for CLI factory
"""

from argparse import ArgumentParser, HelpFormatter
from dataclasses import dataclass, field, asdict
from typing import (
    Optional,
    List, Dict, Any
)

from coaclient.cli.formatters import RawTextArgsHelpFormatter

__all__ = (
    "Parser",
    "SubParser",
    "BaseParser"
)


@dataclass
class BaseParser:
    """
    BaseParser - base parser description class
    """

    @property
    def exclude(self) -> list:
        """
        Get a list of excluded attributes for converting instance to
        the dictionary.
        """
        return list()

    def asdict(
            self,
            exclude: Optional[List[str]] = None,
            **kwargs
    ) -> Dict[str, Any]:
        """
        Convert instance to the dictionary

        :param: exclude - list of excluded attributes (optional). Extended
                          base list of excluded attributes.
        :return: Dict   - dictionary, where the key is the name of instance
                          attributes and value, is the value of these
                          attributes
        """
        if exclude is None:
            exclude = []
        exclude.extend(self.exclude)
        return {
            key: kwargs.get(key, value)
            for key, value in asdict(self).items()
            if (value is not None or key in kwargs) and key not in self.exclude
        }


@dataclass
class Parser(BaseParser):
    """
    Parser - Parsers description class for cli

    +-----------------------+-------------------------------------------------+
    | Parameter             | Description                                     |
    +-----------------------+-------------------------------------------------+
    | prog                  | The name of the program (default: sys.argv[0])  |
    +-----------------------+-------------------------------------------------+
    | usage                 | The string describing the program usage         |
    |                       | (default: generated from arguments added to     |
    |                       |  parser)                                        |
    +-----------------------+-------------------------------------------------+
    | description           | Text to display before the argument help        |
    |                       | (default: none)                                 |
    +-----------------------+-------------------------------------------------+
    | epilog                | Text to display after the argument help         |
    |                       | (default: none)                                 |
    +-----------------------+-------------------------------------------------+
    | parents               | A list of ArgumentParser objects whose          |
    |                       | arguments should also be included               |
    +-----------------------+-------------------------------------------------+
    | formatter_class       | A class for customizing the help output         |
    +-----------------------+-------------------------------------------------+
    | prefix_chars          | The set of characters that prefix optional      |
    |                       | arguments (default: ‘-‘)                        |
    +-----------------------+-------------------------------------------------+
    | fromfile_prefix_chars | The set of characters that prefix files from    |
    |                       | which additional arguments should be read       |
    |                       | (default: None)                                 |
    +-----------------------+-------------------------------------------------+
    | argument_default      | The global default value for arguments          |
    |                       | (default: None)                                 |
    +-----------------------+-------------------------------------------------+
    | conflict_handler      | The strategy for resolving conflicting          |
    |                       | optionals (usually unnecessary)                 |
    +-----------------------+-------------------------------------------------+
    | add_help              | Add a -h/--help option to the parser            |
    |                       | (default: True)                                 |
    +-----------------------+-------------------------------------------------+
    | allow_abbrev          | Allows long options to be abbreviated if the    |
    |                       | abbreviation is unambiguous. (default: True)    |
    +-----------------------+-------------------------------------------------+
    """
    # _pylint: disable=too-many-instance-attributes
    # Disable this warning for dataclass or we can set the value of the
    # `max-attributes` field to more than 7 in the .pylintrc file
    # (` max-attributes` is 7 by default).
    name: Optional[str] = None
    prog: Optional[str] = None
    usage: Optional[str] = None
    description: Optional[str] = None
    epilog: Optional[str] = None
    parents: List[ArgumentParser] = field(default_factory=list)
    formatter_class: HelpFormatter = RawTextArgsHelpFormatter
    prefix_chars: str = '-'
    fromfile_prefix_chars: Optional[str] = None
    argument_default: Optional[str] = None
    conflict_handler: str = 'error'
    add_help: bool = True
    allow_abbrev: bool = True
    help: Optional[str] = None
    subparser: Optional['SubParser'] = None
    defaults: Dict[str, Any] = field(default_factory=dict)
    func: Optional[object] = None
    args: List[str] = field(default_factory=list)

    @property
    def exclude(self) -> list:
        """
        Get a list of excluded attributes in the Parser class for converting
        that instance to the dictionary.
        """
        return ['subparser', 'defaults', 'func', 'args', 'name']


@dataclass
class SubParser(BaseParser):
    """
    SubParser - Subparsers description class for cli

    +--------------+----------------------------------------------------------+
    | Parameter    | Description                                              |
    +--------------+----------------------------------------------------------+
    | title        | title for the sub-cli group in help output; by           |
    |              | default “subcommands” if description is provided,        |
    |              | otherwise uses title for positional arguments            |
    +--------------+----------------------------------------------------------+
    | description  | description for the sub-cli group in help output,        |
    |              | by default None                                          |
    +--------------+----------------------------------------------------------+
    | prog         | usage information that will be displayed with            |
    |              | sub-command help, by default the name of the program     |
    |              | and any positional arguments before the subparser        |
    |              | argument                                                 |
    +--------------+----------------------------------------------------------+
    | parser_class | class which will be used to create sub-cli               |
    |              | instances, by default the class of the current cli       |
    |              | (e.g. ArgumentParser)                                    |
    +--------------+----------------------------------------------------------+
    | action       | the basic type of action to be taken when this argument  |
    |              | is encountered at the command line                       |
    +--------------+----------------------------------------------------------+
    | dest         | name of the attribute under which sub-command name       |
    |              | will be stored; by default None and no value is stored   |
    +--------------+----------------------------------------------------------+
    | required     | Whether or not a subcommand must be provided,            |
    |              | by default False.                                        |
    +--------------+----------------------------------------------------------+
    | help         | help for sub-cli group in help output,                   |
    |              | by default None                                          |
    +--------------+----------------------------------------------------------+
    | metavar      | string presenting available sub-commands in help;        |
    |              | by default it is None and presents sub-commands          |
    |              | in form {cmd1, cmd2, ..}                                 |
    +--------------+----------------------------------------------------------+
    """
    # _pylint: disable=too-many-instance-attributes
    # Disable this warning for dataclass or we can set the value of the
    # `max-attributes` field to more than 7 in the .pylintrc file
    # (` max-attributes` is 7 by default).
    title: Optional[str] = None
    description: Optional[str] = None
    prog: Optional[str] = None
    parser_class: ArgumentParser = ArgumentParser
    action: Optional[str] = None
    dest: Optional[str] = None
    required: Optional[bool] = False
    help: Optional[str] = None
    metavar: Optional[str] = None
    parsers: List['Parser'] = field(default_factory=list)

    @property
    def exclude(self):
        """
        Get a list of excluded attributes in the SubParser class for
        converting that instance to the dictionary.
        """
        return ['parsers', ]



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth2 client command

Command: config
"""

import logging
import os
import re
import time

import requests

from coaclient import oauth2
from coaclient.cli import Parser, SubParser, Arg, Actions
from coaclient.oauth2 import Config
from coaclient.oauth2.utils import validate_input_data
from .exceptions import CoaClientCommandException

__all__ = (
    "add_command",
)


_REGEXP_FILE_NAME = re.compile(r'[^\w\-\_.]')


def add_app(args):
    """
    Adding configuration and credentials for a specific application for
    authorizing in Coursera OAuth2.0 client.
    """
    app_name = args.app
    config_file = args.config if args.config is not None else None
    config = Config.load_from_file(filename=config_file)
    if not config.has_section(app_name) or args.reconfigure is True:
        logging.info(
            "Configuration application \"%s\" for Coursera OAuth2.0 client.",
            app_name
        )
        # Get client_id
        client_id = args.client_id or validate_input_data(
            "Please enter the client id for your application: ",
            empty=False
        )
        # Get client_secret
        client_secret = args.client_secret or validate_input_data(
            "Please enter the client secret for your application: ",
            empty=False
        )
        # Get scopes
        if args.scopes:
            scopes = args.scopes
        else:
            scopes = validate_input_data(
                "Please enter the requested scopes of the app not including "
                "\"view_profile\", separated by whitespace\n(for example: "
                "access_business_api): "
            )

        # Removing app section if exist
        config.remove_section(app_name)
        # Adding new empty app section
        config.add_section(app_name)
        # Adding application credentials to config
        config.set(app_name, "client_id", client_id)
        config.set(app_name, "client_secret", client_secret)
        config.set(app_name, "scopes", "view_profile {scopes}".format(
            scopes=scopes
        ))
        config.set(
            app_name, "token_cache_file", "{app_name}_oauth2_cache.co".format(
                app_name=re.sub(_REGEXP_FILE_NAME, "_", app_name).lower()
            )
        )

        # Save config to file
        config.save(filename=config_file)
        logging.info("Application \"%s\" configured.", app_name)
    else:
        logging.info(
            "Application \"%s\" is already configure and saved to the config "
            "file. You can reconfigure this application if you remove the "
            "config file or use \"%s\" flag when adding configuration for "
            "your application.", app_name, "--reconfigure"
        )


def authorize(args):
    """
    Authorizes Coursera OAuth2.0 client for a specific application
    for using coursera.org API
    """
    if oauth2.build(args.app, args=args).authorizer:
        logging.info("Application \"%s\" authorized.", args.app)
    else:
        logging.error("Something wrong. Application \"%s\" is not "
                      "authorized.", args.app)


def check_auth(args):
    """
    Checking if Coursera OAuth2.0 client connectivity to the coursera.org API
    for a specific application
    """
    profile_url = ("https://api.coursera.org/api/externalBasicProfiles.v1?"
                   "q=me&fields=name")
    response = requests.get(
        profile_url, auth=oauth2.build(args.app, args=args).authorizer
    )

    if response.status_code != requests.codes.ok:  # pylint: disable=no-member
        logging.error('Received response status code %s from the basic '
                      'profile API.', response.status_code)
        logging.debug('Response body: %s', response.text)
        raise Exception('Received response status code {code} from the basic '
                        'profile API.'.format(code=response.status_code))

    response_data = response.json()

    if "elements" in response_data and len(response_data["elements"]) > 0:
        element = response_data["elements"][0]

        if not isinstance(element, dict):
            raise CoaClientCommandException(
                "An invalid data type was received from Coursera OAuth2.0 "
                "API. Type: {type}, Data: {element}".format(
                    type=type(element), element=element
                )
            )

        external_id = element.get("id")

        if external_id is None:
            raise CoaClientCommandException(
                "Could not find the 'external_id' from the response body. "
                "Data: {element}".format(element=element)
            )

        name = element.get("name")

        if name is None:
            raise CoaClientCommandException(
                "Could not find the 'name' from the response body. Data: "
                "{element}".format(element=element)
            )

        logging.info("Name: %s", name)
        logging.info("External ID: %s", external_id)
    else:
        raise CoaClientCommandException(
            "Incorrect data was received from the Coursera OAuth2.0 API. "
            "Data: {response_data}".format(response_data=response_data)
        )


def display_auth_cache(args):
    """
    Output to the screen the state of the authentication cache.

    DEVELOPER NOTE: For debugging authentication issues.

    BEWARE: DO NOT send them to third-party service or via email!!!
    You must keep the tokens secure.
    Treat them as passwords.
    """
    auth = oauth2.build(args.app, args=args)

    token = auth.cache.get('token', '')
    expires = auth.cache.get('expires', 0.0) - time.time()
    refresh = auth.cache.get('refresh', None)
    if not args.no_truncate:
        token = "{}**********{}".format(token[:3], token[-3:])
        if refresh is not None:
            refresh = "{}**********{}".format(refresh[:3], refresh[-3:])

    logging.info("Authorization token: %s", token)
    logging.info(
        "Authorization token is already expired."
        if expires < 0 else
        "Authorization token expires in %.2f seconds",
        expires
    )
    if refresh is not None:
        logging.info("Refresh token: %s", refresh)
    else:
        logging.warning("Refresh token not found.")


def delete(args):
    """
    Delete the application from the configuration if the application exists.
    """
    # oauth2.delete_application(args.app)
    app_name = args.app
    config_file = args.config if args.config is not None else None
    config = Config.load_from_file(filename=config_file)
    if not config.has_section(app_name):
        raise CoaClientCommandException(
            "Configuration for {app_name} application does not exist or "
            "already removed from configuration file.".format(
                app_name=app_name
            )
        )

    choice = validate_input_data(
        "Are you absolutely sure that you want to delete the application "
        "{app_name}? (Y/N): ".format(app_name=app_name)
    ).strip().lower()

    if choice in ("yes", "y"):
        logging.info("Deleting the application \"%s\"", app_name)
        cache_file = os.path.join(
            config.get(config.OAUTH2_SECTION, "token_cache_path"),
            config.get(app_name, "token_cache_file")
        )
        if os.path.exists(cache_file) and os.path.isfile(cache_file):
            os.remove(cache_file)
        config.remove_section(app_name)
        config.save(config_file)
        logging.info("Application \"%s\" was removed", app_name)


def add_command(cli_factory):
    """
    Create config command with command handlers for configure sub commands and
    add to the Coursera's CLI
    """
    config = Parser(
        name="config",
        help="Configure %(prog)s for OAuth2.0 operations",
        subparser=SubParser()
    )
    cli_factory.parser.subparser.parsers.append(config)
    cli_factory.add_arguments(app=Arg(
        flags=("-a", "--app"),
        type=str,
        required=True,
        help="Name of application to configure"
    ), reconfigure=Arg(
        flags=("--reconfigure",),
        action=Actions.STORE_TRUE,
        help="Reconfigure existing application."
    ), no_truncate=Arg(
        flags=("--no-truncate",),
        action=Actions.STORE_TRUE,
        help="[!!! DANGER !!!] Do not truncate the keys. [!!! DANGER !!!].\n"
             "Do that on your own risk and we think you understand that "
             "you do."
    ), client_id=Arg(
        flags=("--client-id",),
        type=str,
        help="Application client id."
    ), client_secret=Arg(
        flags=("--client-secret",),
        type=str,
        help="Application client secret."
    ), scopes=Arg(
        flags=("--scopes",),
        action=Actions.APPEND,
        type=str,
        help="Application scopes. (E.g: view_profile or access_business_api)"
    ))

    # Create sub commands for config command
    # 1. add
    config.subparser.parsers.append(Parser(
        name="add",
        func=add_app,
        help="Adding configuration and credentials for a specific application "
             "for authorizing in Coursera OAuth2.0 client.",
        args=['app', 'reconfigure', 'client_id', 'client_secret', 'scopes']
    ))
    # 2. authorize
    config.subparser.parsers.append(Parser(
        name="authorize",
        func=authorize,
        help="Authorizes Coursera OAuth2.0 client for a specific application"
             " for using coursera.org API",
        args=['app', ]
    ))
    # 3. check-auth
    config.subparser.parsers.append(Parser(
        name="check-auth",
        func=check_auth,
        help="Check Coursera OAuth2.0 client connectivity to the coursera.org "
             "API for a specific application",
        args=['app', ]
    ))
    # 4. display-auth-cache
    config.subparser.parsers.append(Parser(
        name="display-auth-cache",
        func=display_auth_cache,
        help="Output to the screen the state of the authentication cache.\n"
             "BEWARE: DO NOT send them to third-party service or via "
             "email!!!\nYou must keep the tokens secure.\nTreat them as "
             "passwords.",
        args=['app', 'no_truncate']
    ))
    # 5. delete
    config.subparser.parsers.append(Parser(
        name="delete",
        func=delete,
        help="Delete the application from configuration file if the "
             "application exists",
        args=['app', ]
    ))



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coursera OAuth default settings
"""
import os as _os

OAUTH2_CONFIG_DIR = ".coursera"
OAUTH2_CONFIG_PATH = _os.path.join(_os.getenv("HOME", ""), OAUTH2_CONFIG_DIR)
OAUTH2_CONFIG_FILE_NAME = "coaclient.cfg"
OAUTH2_HOSTNAME = "localhost"
OAUTH2_PORT = 9876
OAUTH2_API_ENDPOINT = "https://api.coursera.org"
OAUTH2_BASE_ENDPOINT = "https://accounts.coursera.org/oauth2/v1"
OAUTH2_AUTH_ENDPOINT = "{base_endpoint}/auth".format(
    base_endpoint=OAUTH2_BASE_ENDPOINT
)
OAUTH2_TOKEN_ENDPOINT = "{base_endpoint}/token".format(
    base_endpoint=OAUTH2_BASE_ENDPOINT
)
OAUTH2_VERIFY_TLS = True
OAUTH2_TOKEN_CACHE_PATH = OAUTH2_CONFIG_PATH



# Copyright 2020 Coursera
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Coaclient Args description for CLI factory
"""

from dataclasses import dataclass
from typing import (
    Tuple,
    Union,
    Text,
    Any,
    Type,
    List,
    Optional
)

__all__ = (
    "Actions", "Arg"
)


class Actions:
    """ Actions for cli """
    # _pylint: disable=too-few-public-methods
    # Disable this warning because that is a static class or
    # we can provide .pylintrc file where will
    # setup `min-public-methods` less than 2
    # (`min-public-methods=2` uses by pylint as default).
    STORE: str = 'store'
    STORE_CONST: str = 'store_const'
    STORE_TRUE: str = 'store_true'
    STORE_FALSE: str = 'store_false'
    APPEND: str = 'append'
    APPEND_CONST: str = 'append_const'
    COUNT: str = 'count'
    VERSION: str = 'version'
    HELP: str = 'help'


@dataclass
class Arg:
    """
    Arg - Argument description class for cli

    +-----------+-------------------------------------------------------------+
    | Parameter | Description                                                 |
    +-----------+-------------------------------------------------------------+
    | flags     | Either a name or a list of option strings, e.g. foo or      |
    |           | -f, --foo.                                                  |
    +-----------+-------------------------------------------------------------+
    | action    | The basic type of action to be taken when this argument is  |
    |           | encountered at the command line.                            |
    +-----------+-------------------------------------------------------------+
    | nargs     | The number of command-line arguments that should be         |
    |           | consumed.                                                   |
    +-----------+-------------------------------------------------------------+
    | const     | A constant value required by some action and nargs          |
    |           | selections.                                                 |
    +-----------+-------------------------------------------------------------+
    | default   | The value produced if the argument is absent from the       |
    |           | command line.                                               |
    +-----------+-------------------------------------------------------------+
    | type      | The type to which the command-line argument should be       |
    |           | converted.                                                  |
    +-----------+-------------------------------------------------------------+
    | choices   | A container of the allowable values for the argument.       |
    +-----------+-------------------------------------------------------------+
    | required  | Whether or not the command-line option may be omitted       |
    |           | (optionals only).                                           |
    +-----------+-------------------------------------------------------------+
    | help      | A brief description of what the argument does.              |
    +-----------+-------------------------------------------------------------+
    | metavar   | A name for the argument in usage messages.                  |
    +-----------+-------------------------------------------------------------+
    | dest      | The name of the attribute to be added to the object         |
    |           | returned by parse_args().                                   |
    +-----------+-------------------------------------------------------------+
    """
    # _pylint: disable=too-many-instance-attributes
    # Disable this warning for dataclass or we can set the value of the
    # `max-attributes` field to more than 7 in the .pylintrc file
    # (` max-attributes` is 7 by default).

    flags: Union[Tuple, str]
    help: Text = ""
    action: str = Actions.STORE
    default: Any = None
    type: Optional[Type] = None
    choices: Optional[List[Any]] = None
    metavar: Optional[str] = None
    nargs: Optional[Union[int, str]] = None
    const: Optional[Any] = None
    required: bool = False
    dest: Optional[str] = None
